#!/usr/bin/python

import logging

import dnfile
from dnfile.enums import MetadataTables

from .AbstractLabelProvider import AbstractLabelProvider

LOGGER = logging.getLogger(__name__)


class CilSymbolProvider(AbstractLabelProvider):
    """ Minimal resolver for CIL/DOTNET symbols """

    def __init__(self, config):
        self._config = config
        #addr:func_name
        self._addr_to_func_symbols = {}
        self._func_symbol_to_addr = {}

    def isSymbolProvider(self):
        return True
    
    def decodeSymbolName(self, value):
        """ ensure a proper utf-8 escaped string """
        return value.encode("utf-8").decode("utf-8")

    def update(self, binary_info):
        pe = dnfile.dnPE(data=binary_info.raw_data)
        for row in pe.net.mdtables.MethodDef:
            addr = pe.get_offset_from_rva(row.Rva)
            func_name = self.decodeSymbolName(row.Name.value)
            self._addr_to_func_symbols[addr] = func_name
            self._func_symbol_to_addr[func_name] = addr

    def getSymbol(self, address):
        return self._addr_to_func_symbols.get(address, "")

    def getAddress(self, func_symbol):
        return self._func_symbol_to_addr.get(func_symbol, None)

    def getFunctionSymbols(self):
        return self._addr_to_func_symbols



import string
import struct
from typing import Tuple, Iterator

from smda.common import SmdaFunction

# ported back from our PR to capa v4.0.0
# https://github.com/mandiant/capa/blob/v4.0.0/capa/features/extractors/smda/insn.py


def read_bytes(smda_report, va, num_bytes=None):
    """
    read up to MAX_BYTES_FEATURE_SIZE from the given address.
    """

    rva = va - smda_report.base_addr
    if smda_report.buffer is None:
        raise ValueError("buffer is empty")
    buffer_end = len(smda_report.buffer)
    max_bytes = num_bytes if num_bytes is not None else 0x100
    if rva + max_bytes > buffer_end:
        return smda_report.buffer[rva:]
    else:
        return smda_report.buffer[rva : rva + max_bytes]


def derefs(smda_report, p):
    """
    recursively follow the given pointer, yielding the valid memory addresses along the way.
    useful when you may have a pointer to string, or pointer to pointer to string, etc.

    this is a "do what i mean" type of helper function.

    based on the implementation in viv/insn.py
    """
    depth = 0
    while True:
        if not smda_report.isAddrWithinMemoryImage(p):
            return
        yield p

        bytes_ = read_bytes(smda_report, p, num_bytes=4)
        val = struct.unpack("I", bytes_)[0]

        # sanity: pointer points to self
        if val == p:
            return

        # sanity: avoid chains of pointers that are unreasonably deep
        depth += 1
        if depth > 10:
            return

        p = val


def detect_ascii_len(smda_report, offset):
    if smda_report.buffer is None:
        return 0
    ascii_len = 0
    rva = offset - smda_report.base_addr
    char = smda_report.buffer[rva]
    while char < 127 and chr(char) in string.printable:
        ascii_len += 1
        rva += 1
        char = smda_report.buffer[rva]
    if char == 0:
        return ascii_len
    return 0


def detect_unicode_len(smda_report, offset):
    if smda_report.buffer is None:
        return 0
    unicode_len = 0
    rva = offset - smda_report.base_addr
    char = smda_report.buffer[rva]
    second_char = smda_report.buffer[rva + 1]
    while char < 127 and chr(char) in string.printable and second_char == 0:
        unicode_len += 2
        rva += 2
        char = smda_report.buffer[rva]
        second_char = smda_report.buffer[rva + 1]
    if char == 0 and second_char == 0:
        return unicode_len
    return 0


def read_string(smda_report, offset):
    alen = detect_ascii_len(smda_report, offset)
    if alen > 1:
        return read_bytes(smda_report, offset, alen).decode("utf-8")
    ulen = detect_unicode_len(smda_report, offset)
    if ulen > 2:
        return read_bytes(smda_report, offset, ulen).decode("utf-16")


def extract_strings(f: SmdaFunction) -> Iterator[Tuple[str, int]]:
    """parse string features from the given instruction."""
    for insn in f.getInstructions():
        for data_ref in insn.getDataRefs():
            for v in derefs(f.smda_report, data_ref):
                string_read = read_string(f.smda_report, v)
                if string_read:
                    yield string_read.rstrip("\x00"), insn.offset



import struct
import hashlib
from typing import Iterator

from smda.common.SmdaInstruction import SmdaInstruction


class SmdaBasicBlock:

    smda_function = None
    instructions = None
    picblockhash = None
    opcblockhash = None
    offset = None
    length = None

    def __init__(self, instructions, smda_function=None):
        assert isinstance(instructions, list)
        self.smda_function = smda_function
        if instructions:
            self.instructions = instructions
            self.offset = instructions[0].offset
            self.length = len(instructions)
            self.picblockhash = self.getPicBlockHash()
            self.opcblockhash = self.getOpcBlockHash()

    def getInstructions(self) -> Iterator["SmdaInstruction"]:
        for instruction in self.instructions:
            yield instruction

    def getPicBlockHash(self):
        if self.picblockhash is not None:
            return self.picblockhash
        picblockhash_sequence = self.getPicBlockHashSequence()
        if picblockhash_sequence is not None:
            self.picblockhash = struct.unpack("Q", hashlib.sha256(picblockhash_sequence).digest()[:8])[0]
        return self.picblockhash

    def getPicBlockHashSequence(self):
        """ if we have a SmdaFunction as parent, we can try to generate the PicBlockHash ad-hoc """
        # check all the prerequisites
        if self.smda_function and self.smda_function.smda_report and self.smda_function._escaper and self.smda_function.smda_report.base_addr is not None and self.smda_function.smda_report.binary_size:
            escaped_binary_seqs = []
            for instruction in self.getInstructions():
                escaped_binary_seqs.append(instruction.getEscapedBinary(self.smda_function._escaper, escape_intraprocedural_jumps=True, lower_addr=self.smda_function.smda_report.base_addr, upper_addr=self.smda_function.smda_report.base_addr + self.smda_function.smda_report.binary_size))
            return bytes([ord(c) for c in "".join(escaped_binary_seqs)])
    
    def getOpcBlockHash(self):
        if self.opcblockhash is not None:
            return self.opcblockhash
        opcblockhash_sequence = self.getOpcBlockHashSequence()
        if opcblockhash_sequence is not None:
            self.opcblockhash = struct.unpack("Q", hashlib.sha256(opcblockhash_sequence).digest()[:8])[0]
        return self.opcblockhash

    def getOpcBlockHashSequence(self):
        """ if we have a SmdaFunction as parent, we can try to generate the OpcBlockHash ad-hoc """
        # check all the prerequisites
        if self.smda_function and self.smda_function.smda_report and self.smda_function._escaper:
            escaped_binary_seqs = []
            for instruction in self.getInstructions():
                escaped_binary_seqs.append(instruction.getEscapedToOpcodeOnly(self.smda_function._escaper))
            return bytes([ord(c) for c in "".join(escaped_binary_seqs)])
        
    def getPredecessors(self):
        predecessors = []
        if self.smda_function is not None:
            for frm, to in self.smda_function.blockrefs.items():
                if self.offset in to:
                    predecessors.append(frm)
        return predecessors
    
    def getSuccessors(self):
        successors = []
        if self.smda_function is not None:
            if self.offset in self.smda_function.blockrefs:
                successors.extend(self.smda_function.blockrefs[self.offset])
        return successors

    @classmethod
    def fromDict(cls, block_dict, smda_function=None) -> "SmdaBasicBlock":
        smda_block = cls(None)
        smda_block.smda_function = smda_function
        smda_block.instructions = [SmdaInstruction.fromDict(d, smda_function=smda_function) for d in block_dict]
        return smda_block

    def toDict(self) -> dict:
        return [smda_ins.toDict() for smda_ins in self.instructions]

    def __int__(self):
        return self.offset

    def __str__(self):
        return "0x{:08x}: ({:>4})".format(self.offset, self.length)
